import inspect
import logging
import os
import platform
import time
from contextlib import contextmanager

from komtet_kassa_linux.driver import IFptr

DRIVER_PLATFORM_DIRECTORY = {
    'armv7l': 'linux-armhf',
    'x86_64': 'linux-x64'
}
LIBRARY_PATH = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(IFptr))),
                            DRIVER_PLATFORM_DIRECTORY.get(platform.machine(), 'linux-armhf'))

logger = logging.getLogger(__name__)


class KKTException(Exception):

    def __init__(self, code, description):
        self.code = code
        self.description = description


class KKT:

    def __init__(self, device_path, rent_station=None):
        self.fptr = IFptr(os.path.join(LIBRARY_PATH, "libfptr10.so"))
        self.fptr.setSingleSetting(IFptr.LIBFPTR_SETTING_LIBRARY_PATH, LIBRARY_PATH)
        self.fptr.setSingleSetting(IFptr.LIBFPTR_SETTING_MODEL, str(IFptr.LIBFPTR_MODEL_ATOL_AUTO))

        self.fptr.setSingleSetting(IFptr.LIBFPTR_SETTING_PORT, str(IFptr.LIBFPTR_PORT_USB))
        self.fptr.setSingleSetting(IFptr.LIBFPTR_SETTING_USB_DEVICE_PATH, device_path)

        self.fptr.applySingleSettings()
        self.rent_station = rent_station
        self.device_path = device_path

    def __del__(self):
        del self.fptr

    def driver_version(self):
        return self.fptr.version()

    def _connect(self):
        self.fptr.open()
        while not self.fptr.isOpened():
            logger.info('Device[%s] is busy', self.device_path)
            time.sleep(1)
            self.fptr.open()

        return self

    def _close(self):
        self.fptr.close()

    @contextmanager
    def connect(self):
        try:
            yield self._connect()
        finally:
            self._close()

    @contextmanager
    def query(self, data_type):
        with self.connect() as kkt:
            kkt.fptr.setParam(IFptr.LIBFPTR_PARAM_DATA_TYPE, data_type)
            kkt.fptr.queryData()
            yield kkt.fptr

    def _get_data(self, data_type, tag, param_type='string'):
        with self.connect() as kkt:
            kkt.fptr.setParam(IFptr.LIBFPTR_PARAM_FN_DATA_TYPE, data_type)
            kkt.fptr.fnQueryData()
            if param_type == 'integer':
                return kkt.fptr.getParamInt(tag)
            elif param_type == 'datetime':
                return kkt.fptr.getParamDateTime(tag)

            return kkt.fptr.getParamString(tag)

    def get_reg_info(self, tag, param_type='string'):
        return self._get_data(IFptr.LIBFPTR_FNDT_REG_INFO, tag, param_type)

    def get_fn_info(self, tag, param_type='string'):
        return self._get_data(IFptr.LIBFPTR_FNDT_FN_INFO, tag, param_type)

    @property
    def date_time(self):
        with self.query(IFptr.LIBFPTR_DT_STATUS) as query:
            return query.getParamDateTime(IFptr.LIBFPTR_PARAM_DATE_TIME)

    @property
    def serial_number(self):
        with self.query(IFptr.LIBFPTR_DT_STATUS) as query:
            return query.getParamString(IFptr.LIBFPTR_PARAM_SERIAL_NUMBER)

    @property
    def inn(self):
        return self.get_reg_info(1018)

    @property
    def reg_number(self):
        return self.get_reg_info(1037)

    @property
    def fiscal_drive_id(self):
        return self.get_fn_info(IFptr.LIBFPTR_PARAM_SERIAL_NUMBER)

    @property
    def fn_state(self):
        return self.get_fn_info(IFptr.LIBFPTR_PARAM_FN_FLAGS, 'integer')

    @property
    def ofd_url(self):
        return self.get_reg_info(1046)

    @property
    def organisation(self):
        return self.get_reg_info(1048)

    @property
    def organisation_address(self):
        return self.get_reg_info(1009)

    def _check(self, res):
        if res < 0:
            code, description = self.fptr.errorCode(), self.fptr.errorDescription()
            logger.warning('Printer error: %s [%s]', code, description)
            raise KKTException(code, description)

    @property
    def is_closed_document(self):
        return self.fptr.getParamBool(IFptr.LIBFPTR_PARAM_DOCUMENT_CLOSED)

    def get_last_receipt(self):
        self.fptr.setParam(IFptr.LIBFPTR_PARAM_FN_DATA_TYPE, IFptr.LIBFPTR_FNDT_LAST_RECEIPT)
        self.fptr.fnQueryData()

        return {
            'fiscal_document_number': self.fptr.getParamString(
                IFptr.LIBFPTR_PARAM_DOCUMENT_NUMBER
            ),
            # 'intent': self.fptr.getParamInt(IFptr.LIBFPTR_PARAM_RECEIPT_TYPE),
            'sum': self.fptr.getParamDouble(IFptr.LIBFPTR_PARAM_RECEIPT_SUM),
            'fiscal_id': self.fptr.getParamString(IFptr.LIBFPTR_PARAM_FISCAL_SIGN),
            'time': (self.fptr.getParamDateTime(IFptr.LIBFPTR_PARAM_DATE_TIME)
                     .strftime("%Y-%m-%dT%H:%M:%S"))
        }

    def get_shift_info(self):
        self.fptr.setParam(IFptr.LIBFPTR_PARAM_FN_DATA_TYPE, IFptr.LIBFPTR_FNDT_SHIFT)
        self.fptr.fnQueryData()

        return {
            'session_check': self.fptr.getParamString(IFptr.LIBFPTR_PARAM_RECEIPT_NUMBER),
            'session': self.fptr.getParamString(IFptr.LIBFPTR_PARAM_SHIFT_NUMBER)
        }

    def get_info(self):
        return {
            'serial_number': self.serial_number,
            'rent_station': self.rent_station,
            'fiscal_id': self.fiscal_drive_id,
            'reg_till': self.validity,
            'fn_state': self.fn_state
        }

    @property
    def validity(self):
        return self._get_data(IFptr.LIBFPTR_FNDT_VALIDITY,
                              IFptr.LIBFPTR_PARAM_DATE_TIME, 'datetime')

    @property
    def unconfirmed_fiscal_document_count(self):
        return self._get_data(IFptr.LIBFPTR_FNDT_OFD_EXCHANGE_STATUS,
                              IFptr.LIBFPTR_PARAM_DOCUMENTS_COUNT)

    @property
    def last_document_number(self):
        return self._get_data(IFptr.LIBFPTR_FNDT_LAST_DOCUMENT,
                              IFptr.LIBFPTR_PARAM_DOCUMENT_NUMBER)

    @property
    def fn_warnings_and_errors(self):
        warnings = self.fn_state
        if warnings:
            flags = {
                0: 'Требуется срочная замена ФН',
                1: 'Исчерпан ресурс ФН',
                2: 'Память ФН переполнена',
                3: 'Превышено время ожидания ответа от ОФД',
                7: 'Критическая ошибка ФН'
            }
            return list(map(lambda offset: flags.get(offset),
                            filter(lambda offset: (warnings >> offset) & 1, range(7))))
        return []

    def get_detail_info(self):
        return {
            'serial_number': self.serial_number,
            'fiscal_id': self.fiscal_drive_id,
            'validity': self.validity,
            'errors': self.fn_warnings_and_errors,
            'unconfirmed_document_count': self.unconfirmed_fiscal_document_count,
            'last_document': self.last_document_number,
        }
