import datetime
import logging

from komtet_kassa_linux.driver import IFptr
from komtet_kassa_linux.models import Printer


SHIFT_STATE = CLOSED, OPENED, EXPIRED = (
    IFptr.LIBFPTR_SS_CLOSED, IFptr.LIBFPTR_SS_OPENED, IFptr.LIBFPTR_SS_EXPIRED
)


logger = logging.getLogger(__name__)

AUTOCLOSE_SESSION_TIME = (23, 59)


class ShiftException(Exception):

    def __init__(self, code, description):
        self.code = code
        self.description = code


class Shift:

    def __init__(self, kkt):
        self.kkt = kkt

    @property
    def state(self):
        with self.kkt.query(IFptr.LIBFPTR_DT_SHIFT_STATE) as query:
            return query.getParamInt(IFptr.LIBFPTR_PARAM_SHIFT_STATE)

    @property
    def is_open(self):
        return self.state == OPENED

    @property
    def is_closed(self):
        return self.state == CLOSED

    @property
    def is_expired(self):
        return self.state == EXPIRED

    @property
    def number(self):
        if self.state == CLOSED:
            return

        with self.kkt.query(IFptr.LIBFPTR_DT_SHIFT_STATE) as query:
            return query.getParamInt(IFptr.LIBFPTR_PARAM_SHIFT_NUMBER)

    @property
    def _is_closed_document(self):
        return self.kkt.fptr.getParamBool(IFptr.LIBFPTR_PARAM_DOCUMENT_CLOSED)

    def _close_document(self):
        self._check(self.kkt.fptr.checkDocumentClosed())
        if not self._is_closed_document:
            raise ShiftException(999, 'Shift is not closed')

    def _check(self, res):
        return self.kkt._check(res)

    def open(self, cashier_name=None, cashier_inn=None):
        if self.is_open:
            return
        elif self.is_expired:
            self.close()

        with self.kkt.connect() as kkt:
            if cashier_name and cashier_inn:
                kkt.fptr.setParam(1021, cashier_name)
                kkt.fptr.setParam(1203, cashier_inn)
                self._check(kkt.fptr.operatorLogin())

            self._check(kkt.fptr.openShift())
            self._close_document()

    def close(self):
        if self.is_closed:
            return

        with self.kkt.connect() as kkt:
            kkt.fptr.setParam(IFptr.LIBFPTR_PARAM_REPORT_TYPE, IFptr.LIBFPTR_RT_CLOSE_SHIFT)
            self._check(kkt.fptr.report())
            self._close_document()
            printer = Printer.query.filter_by(serial_number=self.kkt.serial_number).first()
            printer.update(session_closed_at=datetime.datetime.now())

    def reopen(self):
        if self.is_open or self.is_expired:
            self.close()
        self.open()

    def autocheck(self):
        time_now = datetime.datetime.now()
        if (time_now.hour, time_now.minute) == AUTOCLOSE_SESSION_TIME:
            printer = Printer.query.filter_by(serial_number=self.kkt.serial_number).first()
            if (
                not printer.session_closed_at or
                (printer.session_closed_at.hour,
                 printer.session_closed_at.minute) != AUTOCLOSE_SESSION_TIME or
                printer.session_closed_at.date() != time_now.date()
            ):
                self.close()
