import datetime
import logging
import os
import shelve

from komtet_kassa_linux.devices import Receipt, Shift
from komtet_kassa_linux.devices.kkt import KKT, KKTException
from komtet_kassa_linux.devices.receipt import Agent, Supplier
from komtet_kassa_linux.libs.komtet_kassa import POS


logger = logging.getLogger(__name__)


STORE_DIR = 'tmp/km_store/'
os.makedirs(STORE_DIR, exist_ok=True)


class Store:

    def __init__(self, kkt_name):
        self._file_path = STORE_DIR + kkt_name

    def _open_shelve(self):
        return shelve.open(self._file_path)

    def get(self):
        with self._open_shelve() as s:
            return s.get('receipt')

    def save(self, value):
        with self._open_shelve() as s:
            s['receipt'] = value


class BaseKM:

    def __init__(self, printer, rent_station=None):
        self.printer = printer
        self.rent_station = rent_station
        self.store = Store(printer.serial_number)
        self.pos = POS(printer.pos_key, printer.pos_secret, printer.serial_number)

    def get_device_info(self):
        raise NotImplementedError

    def fiscalize_receipt(self, task):
        raise NotImplementedError

    def beat(self, is_only_send_report=False):
        if not self.printer.is_virtual:
            Shift(self.kkt).autocheck()

        device_info = self.get_device_info()
        fiscal_data = None
        if 'error_description' not in device_info:
            receipt = self.store.get()
            if receipt:
                fiscal_data = (receipt.get('__fiscal_data__') or
                               self.fiscalize_receipt(receipt))

        try:
            if is_only_send_report:
                receipt = self.pos.send_report(fiscal_data)
            else:
                receipt = self.pos.get_task(device_info, fiscal_data)
        except AssertionError:
            raise
        except Exception:
            if fiscal_data:
                receipt['__fiscal_data__'] = fiscal_data
            logger.exception("Receipt didn't get")
        else:
            return bool(receipt)
        finally:
            self.store.save(receipt)


class KM(BaseKM):

    def __init__(self, printer, rent_station=None):
        super().__init__(printer, rent_station)
        self.kkt = KKT(printer.devname, rent_station=rent_station)._connect()

    def __del__(self):
        self.kkt._close()

    def get_device_info(self):
        try:
            return self.kkt.get_info()
        except KKTException as exc:
            return {
                'serial_number': self.kkt.serial_number,
                'rent_station': self._rent_station,
                'error_description': str(exc)
            }

    def create_receipt(self, kkt, task):
        receipt = Receipt(kkt, task['intent'])
        if task['cashier'] and task['cashier_inn']:
            receipt.set_cashier(task['cashier'], task['cashier_inn'])
        if task.get('client'):
            receipt.set_client(task['client'].get('inn'), task['client'].get('name'))
        receipt.sno = task['sno']
        receipt.email = task.get('user')

        if 'correction' in task:
            receipt.set_correction_info(
                task['correction']['type'],
                task['correction']['description'],
                datetime.datetime.strptime(task['correction']['date'], "%Y-%m-%d"),
                task['correction']['document']
            )
        else:
            receipt.payment_address = task.get('payment_address')

        for position in task['positions']:
            agent = supplier = None
            if position.get('agent_info'):
                info = position['agent_info']
                agent = Agent(kkt, info['type'])
                if info.get('paying_agent'):
                    agent.set_paying_agent(info['paying_agent']['operation'],
                                           info['paying_agent']['phones'])
                if info.get('receive_payments_operator'):
                    agent.set_receive_payments_operator(
                        info['receive_payments_operator']['phones']
                    )
                if info.get('money_transfer_operator'):
                    agent.set_money_transfer_operator(
                        info['money_transfer_operator']['phones'],
                        info['money_transfer_operator']['name'],
                        info['money_transfer_operator']['address'],
                        info['money_transfer_operator']['inn']
                    )

            if position.get('supplier_info'):
                info = position['supplier_info']
                supplier = Supplier(kkt, info['phones'], info['name'], info['inn'])

            receipt.add_position(
                position['name'], position['price'], position['quantity'],
                position['total'], position['vat']['number'],
                measurement_unit=position.get('measure_name'),
                payment_method=position.get('calculation_method'),
                payment_object=position.get('calculation_subject'),
                agent=agent, supplier=supplier
            )

        for payment in task['payments']:
            receipt.add_payment(payment['sum'], payment['type'])

        return receipt

    def fiscalize_receipt(self, task):
        report = {
            "id": task['id'],
            'inn': self.kkt.inn,
            "kkt_serial_number": self.kkt.serial_number,
            "kkt_reg_number": self.kkt.reg_number,
            "fiscal_drive_id": self.kkt.fiscal_drive_id,

            'ofd_url': self.kkt.ofd_url,

            'sno': str(task.get('sno')),
            'cashier': task.get('cashier'),
            'organisation': self.kkt.organisation,
            'organisation_address': self.kkt.organisation_address
        }

        if task.get('inn') != self.kkt.inn:
            report['error_description'] = 'ИНН чека и ККМ не совпадают'
            return report

        shift = Shift(self.kkt)
        shift.open()

        with self.kkt.connect() as kkt:
            receipt = self.create_receipt(kkt, task)

            try:
                report.update(receipt.fiscalize(task.get('print')))
            except KKTException as exc:
                try:
                    receipt.cansel()
                except KKTException as cansel_exc:
                    if cansel_exc.code != 81:
                        raise
                finally:
                    report['error_description'] = str(exc)

        return report


class VirtualKM(BaseKM):

    def get_device_info(self):
        return {
            'serial_number': self.printer.serial_number,
            'rent_station': self.rent_station,
            'fiscal_id': '0000000000000001',
            'reg_till': datetime.date.today() + datetime.timedelta(days=10),
            'fn_state': 0
        }

    def fiscalize_receipt(self, task):
        return {
            "id": task['id'],
            'inn': task['inn'],
            "kkt_serial_number": self.printer.serial_number,
            "kkt_reg_number": '0000000000000001',
            "fiscal_drive_id": '0000000000000001',

            'ofd_url': "ofdp.platformaofd.ru",

            'sno': str(task.get('sno')),
            'cashier': task.get('cashier'),
            'organisation': task.get('org_name') or '',
            'organisation_address': task.get('org_address') or '',

            'fiscal_document_number': '1',
            'sum': sum([payment['sum'] for payment in task.get("payments")]),
            'fiscal_id': '0000000001',
            'time': datetime.datetime.now(),
            'session_check': '1',
            'session': '1'
        }


def factory_km(printer, rent_station=None):
    if printer.is_virtual:
        return VirtualKM(printer, rent_station=rent_station)

    return KM(printer, rent_station=rent_station)
