import logging
import os
import threading
import time

import psutil

from komtet_kassa_linux.devices import KKT
from komtet_kassa_linux.libs.helpers import scan_atol_ttys
from komtet_kassa_linux.models import Printer

from .km import factory_km
from .observers import DeviceObserver, FileObserver


logger = logging.getLogger(__name__)

LIMIT_USE_MEMORY = 75  # %


def wait(timeout, event=None):
    if not event:
        event = threading.Event()
    return event.wait(timeout)


class RepeatThread(threading.Thread):

    def __init__(self, name, func, interval=1):
        self.func = func
        self.interval = interval
        self._stop_event = threading.Event()
        threading.Thread.__init__(self, name=name)

    def run(self):
        while True:
            repeat_immediately = False
            try:
                repeat_immediately = self.func()
            except Exception:
                logger.exception('Error during a beat')

            interval = self.interval
            if repeat_immediately:
                interval = 0.05

            if wait(interval, self._stop_event):
                # self.func(is_only_send_report=True)
                break

    def stop(self):
        self._stop_event.set()


class KmManager:
    _km_threads = {}

    def __init__(self, db_path, interval=10, rent_station=None):
        self._interval = interval
        self._rent_station = rent_station

        self.file_observer = FileObserver(db_path, self.sync)
        self.device_observer = DeviceObserver(self._connect_device, self._disconnect_device)

    def _start_km(self, printer):
        if printer.serial_number in self._km_threads:
            logger.warn('KM[%s] was already run', printer.serial_number)
            self._stop_km(printer.serial_number)

        km = factory_km(printer, rent_station=self._rent_station)
        km_thread = self._km_threads[printer.serial_number] = RepeatThread(
            printer.serial_number, km.beat, self._interval
        )

        km_thread.start()
        logger.info('Start KM[%s] %s', printer.serial_number, printer.pos_key)
        return km_thread

    def _connect_device(self, devname):
        printer = Printer.query.filter_by(serial_number=KKT(devname).serial_number).first()
        if printer:
            printer.update(devname=devname)
            self._start_km(printer)

    def start(self):
        self.device_observer.start()
        self.file_observer.start()

        # Запуск виртуальных принтеров
        for printer in Printer.query.filter(Printer.is_virtual):
            self._start_km(printer)

        # Запуск подключенных Атолов
        for dev in scan_atol_ttys():
            self._connect_device(dev)

    def _stop_km(self, serial_number, is_wait_complete=True):
        km_thread = self._km_threads.pop(serial_number, None)
        if km_thread:
            km_thread.stop()
            if is_wait_complete:
                km_thread.join()
                logger.info('Stop KM[%s]', serial_number)
        elif Printer.query.filter_by(serial_number=serial_number).first():
            logger.warn('KM[%s] is not found', serial_number)

        return km_thread

    def _disconnect_device(self, devname):
        printer = Printer.query.filter_by(devname=devname).first()
        self._stop_km(printer.serial_number, is_wait_complete=True)
        printer.update(devname=None)

    def stop(self):
        self.device_observer.stop()
        self.file_observer.stop()

        # Отправка сигнала завершения КМ
        stopped_km_threads = {
            serial_number: self._stop_km(serial_number, is_wait_complete=False)
            for serial_number in list(self._km_threads.keys())
        }

        # ожидание завершения всех КМ
        for serial_number, km_thread in stopped_km_threads.items():
            if km_thread:
                km_thread.join()
                logger.info('Stop KM[%s]', serial_number)

        self.device_observer.join()
        self.file_observer.join()

        for printer in Printer.query:
            printer.update(devname=None)

    def sync(self):
        registrated_devices = {
            printer.serial_number: printer
            for printer in Printer.query.filter(Printer.is_online)
        }

        # Останавливаем устройства удаленные из web-интерфейса
        for serial_number in (set(self._km_threads) - set(registrated_devices)):
            self._stop_km(serial_number)

        # Запускем устройства добавленные в web-интерфейсе
        for serial_number in (set(registrated_devices) - set(self._km_threads)):
            self._start_km(registrated_devices[serial_number])

    def loop(self):
        logger.info('Start KMManager')
        self.start()

        process = psutil.Process(os.getpid())
        try:
            while not getattr(self, '_stop_loop', False):
                if process.memory_percent() > LIMIT_USE_MEMORY:
                    logger.warning('Stop KMManager: leak memory')
                    break
                time.sleep(0.5)
        except KeyboardInterrupt:
            pass

        self.stop()
        logger.info('Stop KMManager')
