import logging
import subprocess
import time

import requests
from flask import flash, redirect, render_template, request, url_for

import komtet_kassa_linux
from komtet_kassa_linux.devices import KKT
from komtet_kassa_linux.libs import VIRTUAL_PRINTER_PREFIX
from komtet_kassa_linux.libs.helpers import scan_atol_ttys
from komtet_kassa_linux.libs.htpasswd import HtpasswdFile
from komtet_kassa_linux.libs.komtet_kassa import POS
from komtet_kassa_linux.models import Printer

from . import CONFIG, app


logger = logging.getLogger(__name__)


@app.route('/')
def main():
    return redirect(url_for('devices'))


@app.route('/devices')
def devices():
    return render_template('printers.html', **{
        'printers': Printer.query.all(),
        'connected_devices': [KKT(dev).serial_number for dev in scan_atol_ttys()]
    })


@app.route('/devices/registrate', methods=['GET', 'POST'])
def registrate_printer():
    error = None
    devices = {KKT(dev).serial_number: dev for dev in scan_atol_ttys()}

    if request.method == 'POST':
        pos = POS(request.form['pos_key'], request.form['pos_secret'])
        resp = pos.activate(request.form['serial_number'], CONFIG['rent_station'])
        if resp['success']:
            printer = Printer(devname=devices.get(request.form['serial_number']),
                              **request.form.to_dict()).save()
            logger.info('Add %s', printer)
            return redirect(url_for('devices'))

        error = resp.json()['description']
        logger.warging('Activation error: %s', error)

    active_devices = Printer.query.all()
    devices = filter(lambda serial_number: serial_number not in active_devices, devices.keys())
    return render_template('registrate_printer.html',
                           printers=devices,
                           vprinter_sn=VIRTUAL_PRINTER_PREFIX + str(int(time.time() * 10000000)),
                           error=error)


@app.route('/devices/<string:printer_number>/deactivate')
def deactivate_printer(printer_number):
    printer = Printer.query.get(printer_number)

    pos = POS(printer.pos_key, printer.pos_secret)
    try:
        pos.deactivate()
    except Exception:
        pass
    finally:
        printer.delete()
        logger.info('Remove %s', printer)

    return redirect(url_for('devices'))


@app.route('/security', methods=['GET', 'POST'])
def security():
    if request.method == 'POST':
        ht = HtpasswdFile('.htpasswd')
        ht.update('admin', request.form['password'])
        ht.save()
        return redirect(url_for('devices'))

    return render_template('security.html')


@app.route('/logout')
def logout():
    return redirect('http://out@{}'.format(request.host))


@app.route('/info')
def info():
    latest_info = requests.get(CONFIG['latest_version_info']).json()
    return render_template('info.html', **{
        'version': komtet_kassa_linux.__version__,
        'available_version': latest_info['version']
    })


@app.route('/upgrade')
def upgrade():
    latest_info = requests.get(CONFIG['latest_version_info']).json()
    logger.info('Start update to %s', latest_info['version'])
    subprocess.call(' && '.join([
        '. env/bin/activate',
        'pip install -U ' + latest_info['url'],
        'deactivate',
        'sudo supervisorctl restart rkm:'
    ]), shell=True)
    logger.info('Finish update to %s', latest_info['version'])

    return redirect(url_for('raspberry_km.logout'))


@app.route('/devices/<string:printer_number>')
def printer_details(printer_number):
    printer = Printer.query.get(printer_number)
    if printer and (printer.devname in scan_atol_ttys()):
        kkt_info = KKT(printer.devname).get_detail_info()
        return render_template(
            'printer_details.html',
            printer=printer,
            info=kkt_info,
        )
    else:
        flash('Регистратор не подключен')
        return redirect(url_for('devices'))
